
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#include <serialconfig.h>

#include <cdc162.h>
#include <serjtag.h>

#ifndef ADCD
#define ADCD ADC
#endif

#define AUTO_RANGE_SUPPORT	1
#define ALWAYS_ADEN		1

#ifndef USE_EXTERNAL_REF
  #define USE_EXTERNAL_REF	0
#endif
#ifndef CALIBRATE_REF
  #define CALIBRATE_REF		0
#endif


#define poll    usbcdc_poll


#ifdef ADC_EXTENSION

#define ADC_MAX_CONF	8
static uint8_t adc_conf[ADC_MAX_CONF];
/* result format 
 *
 *  uint16_t: unit 0.078125 mV (10/128 mV)
 *
 *  x1 @ V256 
 *   +-+----------+---+
 *   |1|  10      | 5 |
 *   +-+----------+---+
 *     |          |
 *
 */

#if defined(__AVR_ATmega16U4__) || defined(__AVR_ATmega32U4__)

/* 
 *   AIN1    --  ADC0 PF0
 *   AIN2    --  ADC1 PF1
 *   AIN3    --  ADC4 PF4
 *   AIN4    --  ADC5 PF5
 *   AIN5    --  ADC12 PB5
 *   AIN6    --  ADC13 PB6
 *   AIN7    --  ADC6 (internal: EXT_VOLTAGE)
 *     EXT_VOLTAGE assumed 1.27 v - 2.56v
 *     VCC_VOLTAGE assumed 3.2 v - 5.08v(EXT_VOLTAGE * 4)
 */

#define REFS_EXT	(0<<REFS0)
#define REFS_VCC	(1<<REFS0)
#define REFS_V256	(3<<REFS0)
#define CHAN_ADC0	(0x00)
#define CHAN_ADC1	(0x01)
#define CHAN_ADC4	(0x04)
#define CHAN_ADC5	(0x05)
#define CHAN_ADC6	(0x06)
#define CHAN_ADC7	(0x07)
#define CHAN_ADC12	(0x24) // MUX5 = 1
#define CHAN_ADC13	(0x25) // MUX5 = 1
#define CHAN_V110	(0x1e)
#define CHAN_GND	(0x1f)
#define CHAN_TEMP	(0x27) // MUX5 = 1

#define DELTA_ADC01x1	(0x10) // x1
#define DELTA_ADC10x10	(0x09) // x10
#define DELTA_ADC10x200	(0x0b) // x200

#define DELTA_ADC40x10	(0x28) // x10 MUX5 = 1
#define DELTA_ADC40x40	(0x30) // x40 MUX5 = 1
#define DELTA_ADC40x200	(0x38) // x200 MUX5 = 1

#define DELTA_ADC50x10	(0x29) // x10 MUX5 = 1
#define DELTA_ADC50x40	(0x31) // x40 MUX5 = 1
#define DELTA_ADC50x200	(0x39) // x200 MUX5 = 1

#define DELTA_ADC41x1	(0x14) // x1
#define DELTA_ADC41x10	(0x2c) // x10 MUX5 = 1
#define DELTA_ADC41x40	(0x34) // x40 MUX5 = 1
#define DELTA_ADC41x200	(0x3c) // x200 MUX5 = 1

#define DELTA_ADC51x1	(0x15) // x1
#define DELTA_ADC51x10	(0x2d) // x10 MUX5 = 1
#define DELTA_ADC51x40	(0x35) // x40 MUX5 = 1
#define DELTA_ADC51x200	(0x3d) // x200 MUX5 = 1

const static uint8_t adc_chan_se[8] = 
{
	CHAN_ADC0, CHAN_ADC1, CHAN_ADC4, CHAN_ADC5,
	CHAN_ADC12, CHAN_ADC13, CHAN_ADC6, CHAN_TEMP,
};

const static uint8_t adc_chan_delta[8][4] = {
  { 0            , DELTA_ADC40x10, DELTA_ADC40x40,  DELTA_ADC40x200 } , 
  { DELTA_ADC41x1, DELTA_ADC41x10, DELTA_ADC41x40,  DELTA_ADC41x200 } , 
  { 0            , DELTA_ADC50x10, DELTA_ADC50x40,  DELTA_ADC50x200 } , 
  { DELTA_ADC51x1, DELTA_ADC51x10, DELTA_ADC51x40,  DELTA_ADC41x200 } , 
  { 0            , DELTA_ADC10x10, 0             ,  DELTA_ADC10x200 } , 
  { DELTA_ADC01x1, 0             , 0             , 0                } , 
  { 0            , 0             , 0             , 0                } , 
  { 0            , 0             , 0             , 0                } , 
};

#else
#error "ADC_EXTENSION is not supported\n"
#endif

#if CALIBRATE_REF
uint16_t calib_vcc;
uint16_t calib_v256;
#else
  #define calib_vcc  0
  #define calib_v256 0
#endif

#ifndef VCC_VOLTAGE
  #define VCC_VOLTAGE 3.30
#endif
#ifndef EXT_VOLTAGE
  #define EXT_VOLTAGE 1.29
#endif

static inline uint8_t adc_is_busy() {
	return bit_is_set(ADCSRA,ADSC);
}

static inline uint16_t do_adc_native(uint8_t admux) {
	uint16_t r1,r2,r3,tmp;
	ADMUX = admux;
	bit_set(ADMUX,ADLAR);
	if (bit_is_set(admux,ADLAR)) {
		bit_set(ADCSRB,MUX5);
	} else {
		bit_clear(ADCSRB,MUX5);
	}
#ifdef ALWAYS_ADEN
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r1 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r2 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r3 = ADCD;
	if (r2 > r1) {
		tmp = r1; r1 = r2; r2 =tmp;
	}
	if (r3 > r1) {
		tmp = r1; r1 = r3; r3 =tmp;
	}
	if (r3 > r2) {
		tmp = r2; r2 = r3; r3 =tmp;
	}
	return r2;
#else
	ADCSRA |= (1<<ADEN);
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	/* discard 1st result */
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	ADCSRA &= ~(1<<ADEN);
	return ADCD;
#endif
}

void init_adc() {
	/* setup ADC @ 1/64 clock */
#ifdef ALWAYS_ADEN
	bit_set(ADCSRB,ADHSM);
	ADCSRA = (6<<ADPS0) | (1<<ADEN);
#else
	ADCSRA = (6<<ADPS0);
#endif
	/* digial input disable */
	cli();
	DDRF &=  ~( (1<<PF0)|(1<<PF1)|(1<<PF4)|(1<<PF5)|(1<<PF6)|(1<<PF7));
	PORTF &= ~( (1<<PF0)|(1<<PF1)|(1<<PF4)|(1<<PF5)|(1<<PF6)|(1<<PF7));
	sei();
	DIDR0 = (1<<ADC0D) | (1<<ADC1D) 
		| (1<<ADC4D) |(1<<ADC5D)
		| (1<<ADC6D) |(1<<ADC7D);
	cli();
	PORTB &= ~( (1<<PB5)|(1<<PB6));
	DDRB &=  ~( (1<<PB5)|(1<<PB6));
	sei();
	DIDR2 = (1<<ADC12D) | (1<<ADC13D);
}

#if 0
void term_adc() {
	ADCSRA = 0;
	ADCSRB = 0;
	DIDR0 = 0;
	DIDR1 = 0;
}
#endif

int8_t adc_send(uint8_t size) {
	uint8_t i;
	if (size > ADC_MAX_CONF) size = ADC_MAX_CONF;
	for (i=0; i<size; i++) {
		adc_conf[i] = ext_buf[i];
	}
	return i;
}

uint16_t do_adc_se(uint8_t chan);
int16_t do_adc_delta(uint8_t chan);

uint16_t do_adc_se(uint8_t chan) {
	uint8_t range;
	uint8_t admux;
	uint16_t result;
	uint32_t v32;
	range = chan & ADC_RANGE_MASK;
	chan &= ADC_CHAN_MASK;
	admux = adc_chan_se[chan];
#if USE_EXTERNAL_REF
	admux |= REFS_EXT;
	result = do_adc_native(admux);
	v32 = result;
	v32 = v32 * (uint16_t)((EXT_VOLTAGE/2.56) * 0x8000);
	result = v32 >> (1 + 15);
#else
	if (range == ADC_RANGE_V512) {
		admux |= REFS_VCC;
	} else {
		admux |= REFS_V256;
	}
	result = do_adc_native(admux);
	v32 = result;
#if CALIBRATE_REF
	if ((chan == ADC_CHAN_VCC) && (result & 0xc000)) {
	    if (range == ADC_RANGE_V512) {
		calib_vcc = (uint32_t)(0x80000000 * EXT_VOLTAGE/2.56) /result;
	   } else {
		calib_v256 = (uint32_t)(0x80000000 * EXT_VOLTAGE/2.56) /result;
	   }
	}
#endif
	if (range == ADC_RANGE_V512) {
		if (calib_vcc) {
			v32 = v32 * calib_vcc;
		} else {
			v32 = v32 * (uint32_t)(0x8000 * VCC_VOLTAGE/2.56);
		}
		result = v32 >> (1 + 15);
	} else {
		if (calib_v256) {
			v32 = v32 * calib_v256;
			result = v32 >> (1 + 15);
		} else {
			result = result >> 1;
		}
	}
#endif
	return result;
}

int16_t do_adc_delta(uint8_t chan) {
	uint16_t result;
	uint8_t range;
	int32_t v32;
	uint8_t auto_mode = 0;
	uint8_t admux;
	const uint8_t *p;
	range = chan & ADC_RANGE_MASK;
	if (range == ADC_RANGE_AUTO) {
		auto_mode = 1;
		range = ADC_RANGE_V256;
	} else if (range == ADC_RANGE_V512) {
		range = ADC_RANGE_V256;
	}
#if AUTO_RANGE_SUPPORT
range_down:
	p = &(adc_chan_delta[(chan & ADC_CHAN_MASK) - ADC_CHAN_DELTA][0]);
#endif
	if (range == ADC_RANGE_V256) {
		admux = p[0];
		if (auto_mode && (admux == 0)) {
			range = ADC_RANGE_V0256;
			goto range_down;
		}
	} else if (range == ADC_RANGE_V0256) { // x10
		admux = p[1];
		if (auto_mode && (admux == 0)) {
			range = ADC_RANGE_V0064;
			goto range_down;
		}
	} else if (range <= ADC_RANGE_V0064) { // x40
		admux = p[2];
		if (auto_mode && (admux == 0)) {
			range = ADC_RANGE_V00128;
			goto range_down;
		}
	} else {
		admux = p[3];
	}
#if USE_EXTERNAL_REF
	admux |= REFS_EXT;
#else
	admux |= REFS_V256;
#endif
	result = do_adc_native(admux);
	v32 = (uint16_t)result;
	if (range == ADC_RANGE_V256) {
#if AUTO_RANGE_SUPPORT
		if (auto_mode && ((result & 0xf000) == 0) && p[1]) {
			range = ADC_RANGE_V0256;
			goto range_down;
		}
		if (auto_mode && ((result & 0xfc00) == 0) && p[3]) {
			range = ADC_RANGE_V0064;
			goto range_down;
		}
#endif
	} else if (range == ADC_RANGE_V0256) {
#if AUTO_RANGE_SUPPORT
		if (auto_mode && ((result & 0xe000) == 0) && p[2]) {
			range = ADC_RANGE_V0064;
			goto range_down;
		}
#endif
		v32 = v32 * (uint32_t)((0.256/2.56)*0x10000);
		result = v32 >> 16;
	} else if (range <= ADC_RANGE_V0064) {
#if AUTO_RANGE_SUPPORT
		if (auto_mode && ((result & 0xe000) == 0) && p[3]) {
			range = ADC_RANGE_V00128;
			goto range_down;
		}
#endif
		v32 = v32 * (uint32_t)((0.064/2.56)*0x10000);
		result = v32 >> 16;
	} else {
		v32 = v32 * (uint32_t)((0.00128/2.56)*0x10000);
		result = v32 >> 16;
	}
#if USE_EXTERNAL_REF
	v32 = result;
	v32 = v32 * (uint16_t)((EXT_VOLTAGE/2.56) * 0x8000);
	result = v32 >> (1 + 15);
#else
	if (calib_v256) {
		v32 = result;
		v32 = v32 * calib_v256;
		result = v32 >> (1 + 15);
	} else {
		result = result >> 1;
	}
#endif
	return (int16_t)result;
}

int8_t adc_recv(uint8_t size) {
	uint8_t i;
	uint16_t *p = (uint16_t *)ext_buf;
	size >>= 1;

	for (i=0; i<size; i++,p++) {
#if 1
	    if (adc_conf[i] & ADC_CHAN_DELTA) {
		    *p = do_adc_delta(adc_conf[i]);
	    } else {
		    *p = do_adc_se(adc_conf[i]);
            }
#else
		    //*p = temp_val();
		    //*p = adc_conf[i];
		    //*p = vcc_val();
		    //*p = do_adc_se(ADC_CHAN_TEMP);
		    //*p = do_adc_se(ADC_CHAN_VCC | ADC_RANGE_V512);
#endif
	}
	return (i<<1);
}

#endif
